/*
 * This program is run by the process cache wrapper procedure to notify
 * parent procedure that is ready for the next command and then wait for
 * the parent to request the next task.
 *
 * Usage:
 *     await_task req_mbx report_mbx timeout prev_status prefix
 *
 *  where:
 *	req_mbx		Name of request mailbox, we read next task request
 *			from parent.
 *
 *	report_mbx	Name of report mailbox, we write linkdiscon message
 *			to this mailbox to notify parent we will be reading
 *			from the request mailbox.
 *
 *	timeout		Decimal integer that represents the max seconds we
 *			will wait for a task request.
 *
 *	prev_status	Decimal integer that represents the previous
 *			task's completion status, becomes part of linkdiscon
 *			message.
 *
 *	prefix		Prefix for DCL symbols that receive task request
 *			arguments:
 *			    'prefix'_IN	   sys$input for task.
 *			    'prefix'_OUT   sys$output for task.
 *			    'prefix'_ERR   sys$error for task.
 *			    'prefix'_CMD   program/command to execute.
 *
 *  The program uses the directional features of mailboxes to detect
 *  if the parent exits.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <descrip.h>			/* VMS string descriptors */
#include <ssdef.h>			/* VMS sys. service return codes */
#include <dvidef.h>			/* VMS $GETDVI */
#include <devdef.h>			/* VMS device characteristics */
#include <iodef.h>			/* VMS QIO definitions */
#include <cmbdef.h>			/* VMS mailbox item codes */
#include <jpidef.h>			/* VMS job/process information */
#include <lckdef.h>			/* VMS lock manager */
#include <lnmdef.h>

int SYS$ASSIGN(), SYS$QIOW(), LIB$SET_SYMBOL(), SYS$ENQW(), LIB$GETJPI();
int SYS$DEQ(), LIB$GETDVI(), SYS$DASSGN(), SYS$TRNLNM(), LIB$SET_LOGICAL();
int LIB$SET_SYMBOL(), LIB$GET_SYMBOL();
/*
 * Structure definitions.
 */
struct mbx_iosb {
    unsigned short status, count;
    long pid;
}; 
struct mbx_message {
    unsigned short type;		/* message type */
    unsigned short unit;		/* not used */
    int finalsts;			/* exit status */
    char data[248];
};
/*
 * Translate a logical name.
 */
static int translate_logical ( char *name, char *table, char *equiv,
	int equiv_size, int *length, int *attributes )
{
    int status, attr;
    static $DESCRIPTOR(name_dx,"");
    static $DESCRIPTOR(table_dx,"");
    struct { short length, code; char *buffer; int *retlen; } item[5];

    name_dx.dsc$a_pointer = name;
    name_dx.dsc$w_length = strlen(name_dx.dsc$a_pointer);
    table_dx.dsc$a_pointer = table;
    table_dx.dsc$w_length = strlen(table_dx.dsc$a_pointer);

    item[0].length = equiv_size-1; item[0].code = LNM$_STRING;
    item[0].buffer = equiv; item[0].retlen = length;
    *length = 0;		/* zero high byte */
    item[1].length = sizeof(attributes);
    item[1].code = LNM$_ATTRIBUTES;
    item[1].buffer = (char *) attributes;
    item[1].retlen = (int *) 0;
    item[2].length = item[2].code = 0;		/* end list */
    attr = LNM$M_CASE_BLIND;

    status = SYS$TRNLNM ( &attr, &table_dx, &name_dx, 0, item );
    if ( (status&1) == 1 ) equiv[*length] = '\0';
    return status;
}
/***********************************************************************/
/* Assign channel to mailbox and read environment variables.
 */
static int transfer_environment ( struct dsc$descriptor_s *source, 
	char *prefix )
{
    int channel, i, status, code, log_len, logchar, use_log_env, devchar;
    int use_sym_env, prefix_size, prefix_table;
    char log_name[256];
    struct mbx_iosb iosb;
    static $DESCRIPTOR(name_dx,"");
    static $DESCRIPTOR(value_dx,"");
    static $DESCRIPTOR(table_dx,"APACHE_ENV");
    static char sym_prefix[256];
    static $DESCRIPTOR(sym_prefix_dx,sym_prefix);
    /*
     * determine style of environment variables: logical name or DCL or both
     */
    use_log_env = use_sym_env = 0;
    if ( 1&translate_logical(table_dx.dsc$a_pointer, "LNM$PROCESS_DIRECTORY",
	log_name, sizeof(log_name), &log_len, &logchar ) ) {
	/*
	 * Logical name apache_env exists, see if it is a table.
	 */
	if ( (logchar&LNM$M_TABLE) != 0 ) use_log_env = 1;
    }
    sym_prefix_dx.dsc$w_length = sizeof(sym_prefix)-1;
    prefix_size = prefix_table = 0;
    if ( 1&LIB$GET_SYMBOL( &table_dx, &sym_prefix_dx, &prefix_size,
	&prefix_table ) ) {
	use_sym_env = 1;
	sym_prefix[prefix_size] = '\0';
    }
    /*
     * Assign channel to mailbox and verify device type.
     */
    channel = 0;
    status = SYS$ASSIGN ( source, &channel, 0, 0, CMB$M_READONLY );
    if ( (status&1) == 0 ) return status;
    code = DVI$_DEVCHAR;
    devchar = 0;
    status = LIB$GETDVI ( &code, &channel, 0, &devchar, 0, 0 );
    if ( (status&1) == 0 ) return status;
    if ( (devchar&DEV$M_MBX) == 0 ) return SS$_BADPARAM;
    /*
     * Read the mailbox messages.
     */
    while ( status&1 ) {
	char msg[512], *name, *value;

	status = SYS$QIOW ( 0, channel, IO$_READVBLK|IO$M_WRITERCHECK, &iosb,
	    0, 0, msg, sizeof(msg)-1, 0, 0, 0, 0 );
	if ( (status&1) == 1 ) {
	    /*
	     * Do clean exit if end of file message see.  Otherwise return
	     * error code on error.
	     */
	    if ( iosb.status == SS$_ENDOFFILE ) break;
	    status = iosb.status;
	    if ( (status&1) == 0 ) break;
	    /*
	     * Message contains definition for 1 or more environment vars.
	     */
	    msg[iosb.count] = '\0';
	    name = msg;
	    for ( i = 0; i <= iosb.count; i++ ) if ( msg[i] == '\0' ) {
		/*
		 * Parse the string as 'name=value'
		 */
		value = strchr ( msg, '=' );
		if ( value ) {
		    /*
		     * Make string descriptors for name and value.
		     */
		    *value++ = '\0';
		    name_dx.dsc$a_pointer = name;
		    name_dx.dsc$w_length = strlen(name_dx.dsc$a_pointer);
		    value_dx.dsc$a_pointer = value;
		    value_dx.dsc$w_length = strlen(value_dx.dsc$a_pointer);
		    /*
		     * Set string or value as appropriate.
		     */
		    if ( use_log_env && *value ) {
			status = LIB$SET_LOGICAL ( &name_dx, &value_dx,
				&table_dx, 0, 0 );
		    }
		    if ( use_sym_env && 
			((prefix_size+name_dx.dsc$w_length) < 256) ) {
			/*
			 * Construct symbol name as prefix//name.
			 */
			sprintf ( &sym_prefix[prefix_size], "%s", name );
			name_dx.dsc$a_pointer = sym_prefix;
			name_dx.dsc$w_length += prefix_size;

			status = LIB$SET_SYMBOL ( &name_dx, &value_dx,
			    &prefix_table );
		    }
		}
		/*
		 * Update for next null in string.
		 */
		name = &msg[i+1]; 
	    }
	    /* printf ( " (env l=%d): '%s'\n", iosb.count, msg ); */
	}
    }
    SYS$DASSGN ( channel );
    return 1;
}
/***********************************************************************/
/* Parse mailbox message sent by parent and generate DCL symbols.
 */
static char *sym_base[] = { "_OUT", "_IN", "_ERR", "_CMD", "_ENV", (char *) 0};

static int process_request ( 
	long sender, 			/* PID of process that wrote req. */
	char *request, 			/* Request message */
	int reqlen, 			/* Length written */
	char *prefix )			/* Prefix */
{
    int status, length, start, i, j, table;
    static char symname[64];
    static $DESCRIPTOR(symname_dx,symname);
    static $DESCRIPTOR(symval_dx,"");
    /*
     * Break the request buffer into NUL-delimited strings and assign
     * to successive symbols in the symbol list.
     */
    start = 0;
    table=1;
    for ( j = i = 0; i < reqlen; i ++ ) if ( request[i] == '\0' ) {
	/*
	 * Found delimiter for next item.
	 */
	length = i - start;
	sprintf ( symname, "%s%s", prefix, sym_base[j] );
	symname_dx.dsc$w_length = strlen ( symname );
	symval_dx.dsc$a_pointer = &request[start];
	symval_dx.dsc$w_length = length;

	status = LIB$SET_SYMBOL ( &symname_dx, &symval_dx, &table );
	if ( (status&1) == 0 ) {
	    fprintf(stderr,"Error creating DCL symbol %s: %d\n", symname,
		status );
	    return status;
	}
	/*
	 * setup for next.
	 */
	start = i + 1;		/* start of next value */
	j++;			/* next table position */
	/*
	 * Check for end of symbol list.
	 */
	if ( !sym_base[j] ) {
	    /*
	     * If the current symbol value is non-null, it is the name
	     * of the mailbox to transfer the environment array.
	     */
	    if ( length > 0 ) {
		status = transfer_environment ( &symval_dx, prefix );
	    } else status = 1;
	    return status;
	}
    }
    /*
     * Falling out of loop means we ran out of strings in request before
     * we ran out of symbols to set.
     */
    symval_dx.dsc$a_pointer = "";
    symval_dx.dsc$w_length = 0;		/* set null string */
    for ( ; sym_base[j]; j++ ) {
	sprintf ( symname, "%s%s", prefix, sym_base[j] );
	symname_dx.dsc$w_length = strlen ( symname );
	status = LIB$SET_SYMBOL ( &symname_dx, &symval_dx, &table );
	if ( (status&1) == 0 ) {
	    fprintf(stderr,"Error creating DCL symbol %s: %d\n", symname,
		status );
	    return status;
	}
    }
    return 3;
}
/***********************************************************************/
/* Acquire and release the process exit lock.  If this is the second
 * time this process has run this program, the lock won't be acuired
 * (thus stalling the program) until the application on the parent has
 * retrieved the previous completion status from the report mailbox.
 */
static char master_lock_name[32], child_lock_name[32];
static $DESCRIPTOR(master_lock_dx,master_lock_name);
static $DESCRIPTOR(child_lock_dx,child_lock_name);

static int sync_with_exit_lock ( )
{
    int status, code;
    long parent_pid, child_pid;
    long parent_lksb[2], child_lksb[2];
    /*
     * Get our PID and that of parent and generate lock names.
     */
    code = JPI$_OWNER;
    child_pid = 0;
    status = LIB$GETJPI ( &code, &child_pid, 0, &parent_pid, 0, 0 );
    sprintf ( master_lock_name, "PROCESS_CACHE_%x", parent_pid );
    master_lock_dx.dsc$w_length = strlen ( master_lock_dx.dsc$a_pointer );
    sprintf ( child_lock_name, "EXIT_%x", child_pid );
    child_lock_dx.dsc$w_length = strlen ( child_lock_dx.dsc$a_pointer );
    /*
     * Get the locks.
     */
    status = SYS$ENQW ( 0, LCK$K_CRMODE, parent_lksb, 0, &master_lock_dx,
	0, 0, 0, 0, 0, 0 );
    if ( (status&1) == 1 ) status = parent_lksb[0];
    if ( (status&1) == 1 ) {
	status = SYS$ENQW ( 0, LCK$K_PWMODE, child_lksb, 0, &child_lock_dx,
	    parent_lksb[1], 0, 0, 0, 0, 0 );
	if ( (status&1) == 1 ) status = child_lksb[0];
	if ( (status&1) == 1 ) {
	    SYS$DEQ ( child_lksb[1], 0, 0, 0 );
	}
    if ( (status&1) == 0 ) fprintf(stderr, "Error getting lock %s: %d\n",
	child_lock_name, status );
	SYS$DEQ ( parent_lksb[1], 0, 0, 0 );
    }
    if ( (status&1) == 0 ) fprintf(stderr, "Error getting lock %s: %d\n",
	master_lock_name, status );

    return status;
}
/***********************************************************************/
/* Main program.
 */
int main ( int argc, char **argv )
{
    int status, timeout, last_status, req_chan, report_chan;
    char *prefix;
    static $DESCRIPTOR(req_mbx_dx,"");
    static $DESCRIPTOR(report_mbx_dx,"");
    struct mbx_iosb iosb;
    struct mbx_message report;
    char request[512];
    /*
     * verify/decode command line arguments.
     */
    if ( argc != 6 ) {
	fprintf(stderr,"Bad command line, usage:\n  %s %s\n",
	    argv[0], " req_mbx report_mbx timeout prev_status prefix" );
	return SS$_BADPARAM;
    }
    req_mbx_dx.dsc$a_pointer = argv[1];
    req_mbx_dx.dsc$w_length = strlen ( req_mbx_dx.dsc$a_pointer );
    report_mbx_dx.dsc$a_pointer = argv[2];
    report_mbx_dx.dsc$w_length = strlen ( report_mbx_dx.dsc$a_pointer );
    timeout = atoi ( argv[3] );
    last_status = atoi ( argv[4] );
    prefix = argv[5];
   /*
    * Assign channels to mailboxes, we only read from request mailbox and
    * only write to report mailbox.
    */
    req_chan = 0;
    report_chan = 0;
    status = SYS$ASSIGN ( &report_mbx_dx, &report_chan, 0, 0, CMB$M_WRITEONLY);
    if ( (status&1) == 0 ) {
	return status;
    }
    status = SYS$ASSIGN ( &req_mbx_dx, &req_chan, 0, 0, CMB$M_READONLY );
    if ( (status&1) == 0 ) {
	return status;
    }
    /*
     * Send message to report mailbox.   Since channel is already assigned
     * to request mailbox, parent can immediate write next request without
     * getting a NOREADER error.
     */
    report.type = SS$_LINKDISCON;
    report.unit = 0;
    report.finalsts = last_status;
    status = SYS$QIOW ( 0, report_chan, IO$_WRITEVBLK|IO$M_READERCHECK,
	&iosb, 0, 0, &report, sizeof(report)-sizeof(report.data), 0, 0, 0, 0 );
    if ( status&1 ) status = iosb.status;
    if ( (status&1) == 0 ) {
	if ( status != SS$_NOREADER ) {
	    fprintf(stderr,"Error writing to report mailbox: %d\n", status);
	}
	return status;
    }
    /*
     * set timer that will kill read after timeout seconds.
     */
    /*
     * Wait for message, detect if request mailbox loses the writer.
     */
    status = sync_with_exit_lock();
    for ( ; ; )  {
        status = SYS$QIOW ( 0, req_chan, IO$_READVBLK|IO$M_WRITERCHECK, &iosb,
	    0, 0, request, sizeof(request), 0, 0, 0, 0 );
        if ( (status&1) == 1 ) status = iosb.status;
        if ( (status&1) == 1 ) {
	    /*
	     * Got request parse.
	     */
	    status = process_request ( iosb.pid, request, iosb.count, prefix );
	    break;
        } else if ( status == SS$_ENDOFFILE ) {
	    /*
	     * Ignore extraneous end-of-files
	     */
	    continue;
        }
	break;
    }
    return status;
}
